/***********************************************************************/
/*
  r.rifs   in $GIS/src.contrib/MWL

  *****
 
  Mark Lake  12/4/99

  University College London
  Institute of Archaeology
  31-34 Gordon Square
  London.  WC1H 0PY

  Email: mark.lake@ucl.ac.uk

  *****

  PURPOSE 

  1) Produce a raster map and/or graphical image of a fractal specified
  in terms of a random iterated function system.
 
  *****

  METHOD

  1) Iteratively applies T transformations W1, W2,...,WT each with
  probability P1, P2,...PT.  Each transformation W is specified
  according to the matrix

  a, b | e
  c, d | f

  such that if (x,y), (x1,y1),...,(xN,yN) are the set of N transformed
  points then x1 = ax + by + e and y1 = cx + dy + f, etc.

  Note that a,b,c and d provide for linear transformations and e and f
  provide for additional translation (resulting in an affine linear
  transformation).

  2) If the user requests the probablilities P1,...,PT are calculated as

           max (d, |det Wi|)
  Pi =  ----------------------, i = 1,...,T
         k=T
        SUM   max (d, |det Wk|)         
         k=1

  where |det Wi| is the absolute value of ad-bc and d is a small value
  to replace Pi that would otherwise be too close to zero.

  3) For more information see Peitgen, Jurgens and Saupe, 1992, "Chaos
  and Fractals: New Frontiers of Science", Berlin: Springer Verlag,
  chapters 5 & 6.

  *****

  FILES

  1) main.c - algorithm for ifs.

  2) segment_file.c - functions to segment large files for
  random input-output.

  3) display.h - interface to GRASS DISPLAYLIB routines that 
  do not have prototypes in gis.h          

  *****

  PORTABILITY

  1) Machines with a shortage ofRAM might need a higher value for 
  kFRACT in segment_file.c                                             */

/***********************************************************************/


#define MAIN
#define kMAX_TRANSFORMS 10
#define kSMALL_PROB 0.01

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "gis.h"
#include "colors.h"
#include "display.h"
#include "segment.h"
#ifdef ANSIC
#define  "SEGMENTH"
#include "segment_file.c"
#endif

enum parameter {a, b, c, d, e, f};

main (argc, argv)
int argc;
char **argv;
{
  char out_mapset [50];
  char out_seg_name [128];
  int  out_map_fd;
  int  out_seg_fd;
  SEGMENT outSEG;
  char data_name [50];
  char temp [128];
  int projection;
  int array_south, array_north;
  int array_east, array_west;
  int nrows, ncols, row, left;
  int screenx [1], screeny [1];
  CELL *cell;
  CELL val;
  unsigned int rand_seed;
  long int its;
  long int i;
  int count, no_transforms;
  int wdth;
  int n;
  int t;
  float fn [6][kMAX_TRANSFORMS];
  float prob [kMAX_TRANSFORMS];
  float cum_prob [kMAX_TRANSFORMS];
  float determinant [kMAX_TRANSFORMS];
  float r;
  float w;
  float x, xn, y, yn;
  float sum_determinants;
  int ploty;
  char *current_mapset;
  struct History history;
  struct Option *output, *transforms, *iterations, *seed, *color, *width;
  struct Option *opta, *optb, *optc, *optd, *opte, *optf;
  struct Option *optp;
  struct Flag *display, *calc, *overwrite, *silent, *terse;


  /***********************************************************************/
  /* Preliminary things                                                  */
  /***********************************************************************/

  /* Initialize the GIS calls and sort out error handling */
  
  G_gisinit(argv[0]) ;
  G_sleep_on_error (0);


  /* Define the options */
  
  output = G_define_option ();
  output->key = "output";
  output->type = TYPE_STRING;
  output->required = NO;
  output->gisprompt = "new,cell,raster";
  output->description = "Output map";
  transforms = G_define_option ();
  transforms->key = "transforms";
  transforms->type = TYPE_INTEGER;
  transforms->required = YES;
  transforms->description = "Number of transformations (max=10) ";
  iterations = G_define_option ();
  iterations->key = "iterations";
  iterations->type = TYPE_INTEGER;
  iterations->required = NO;
  iterations->answer = "5000";
  iterations->description = "Number of iterations";
  seed = G_define_option ();
  seed->key = "seed";
  seed->required = NO;
  seed->type = TYPE_INTEGER;
  seed->answer = "1";
  seed->description = "Seed for random number generator (min=0, max=32767)";
  color = G_define_option ();
  color->key = "color";
  color->required = NO;
  color->answer = "magenta";
  color->type = TYPE_STRING;
  color->description = "Color for display on monitor ";
  color->options = D_COLOR_LIST;
  width = G_define_option ();
  width->key = "width";
  width->required = NO;
  width->type = TYPE_INTEGER;
  width->description = "Width of fractal (min=0, max=32767) ";
  optp = G_define_option ();
  optp->key = "p";
  optp->required = NO;
  optp->type = TYPE_DOUBLE;
  optp->description = "Probs. for transformations ";
  optp->multiple = YES;
  opta = G_define_option ();
  opta->key = "a";
  opta->required = YES;
  opta->type = TYPE_DOUBLE;
  opta->description = "Parameter a ";
  opta->multiple = YES;
  optb = G_define_option ();
  optb->key = "b";
  optb->required = YES;
  optb->type = TYPE_DOUBLE;
  optb->description = "Parameter b ";
  optb->multiple = YES;
  optc = G_define_option ();
  optc->key = "c";
  optc->required = YES;
  optc->type = TYPE_DOUBLE;
  optc->description = "Parameter c ";
  optc->multiple = YES;
  optd = G_define_option ();
  optd->key = "d";
  optd->required = YES;
  optd->type = TYPE_DOUBLE;
  optd->description = "Parameter d";
  optd->multiple = YES;
  opte = G_define_option ();
  opte->key = "e";
  opte->required = YES;
  opte->type = TYPE_DOUBLE;
  opte->description = "Parameter e ";
  opte->multiple = YES;
  optf = G_define_option ();
  optf->key = "f";
  optf->required = YES;
  optf->type = TYPE_DOUBLE;
  optf->description = "Parameter f ";
  optf->multiple = YES;
  calc = G_define_flag ();
  calc->key = 'c';
  calc->description = "Calculate probabilities ";
  display = G_define_flag ();
  display->key = 'd';
  display->description = "Display in monitor ";
  overwrite = G_define_flag ();
  overwrite->key = 'o';
  overwrite->description = "Overwrite output file if it exists ";
  silent = G_define_flag ();
  silent->key = 's';
  silent->description = 
    "Run silently ";
  terse = G_define_flag ();
  terse->key = 't';
  terse->description = 
    "Terse output only (i.e. only confirm parameters) ";

  if (G_parser(argc, argv))
    exit (-1);


  /* Silent implys terse */

  if (silent->answer)
    terse->answer = 1;


  /* Check that database is UTM or x-y referenced */
  
  projection = G_projection ();
  if (projection != PROJECTION_UTM && projection != PROJECTION_XY)
    G_fatal_error ("Can't use database\n\t\t- not UTM or X-Y");
  

  /* If display on monitor required setup graphics window */
  
  if (display->answer)
    {
      R_open_driver ();
      D_setup(0);
      R_standard_color (D_translate_color (color->answer));
    }
    

  /* If required open output map for writing */
  
  if (output->answer != NULL)
    Segment_named_outfile (output->answer, out_mapset,
			      &out_map_fd, out_seg_name,
			      &out_seg_fd, &outSEG,
			      overwrite->answer, terse->answer);


  /* Calculate array limits */
  
  array_south = G_window_rows () - 1;
  array_north = 0;
  array_west = 0;
  array_east = G_window_cols () - 1;
  nrows = G_window_rows ();
  ncols = G_window_cols ();
  cell = G_allocate_cell_buf ();


  /* Get the numeric parameters */
  
  sscanf (iterations->answer, "%ld", &its);
  sscanf (transforms->answer, "%d", &no_transforms);
  if (no_transforms > kMAX_TRANSFORMS)
    G_fatal_error ("Too many transformations ");
  sscanf (seed->answer, "%d", &rand_seed);
  if (width->answer != NULL)
    sscanf (width->answer, "%d", &wdth);

  
  /* Get transformation parameters */
  
  for (t = 0; t < no_transforms; t++)
    {
      if (opta->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
      else
	sscanf (opta->answers[t], "%f", &fn[a][t]);
      if (optb->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
      else
	sscanf (optb->answers[t], "%f", &fn[b][t]);
      if (optc->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
       else
	 sscanf (optc->answers[t], "%f", &fn[c][t]);
      if (optd->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
      else
	sscanf (optd->answers[t], "%f", &fn[d][t]);
      if (opte->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
      else
	sscanf (opte->answers[t], "%f", &fn[e][t]);
      if (optf->answers[t] == NULL)
	G_fatal_error ("Too few transformations ");
      else
	sscanf (optf->answers[t], "%f", &fn[f][t]);
    }


  /***********************************************************************/
  /* Read or calculate transformation probabilities                      */
  /***********************************************************************/

  /* Get probabilities if provided... */
  
  if (!calc->answer)
    {
      count = 0;
      while (optp->answers[count] != NULL)
	{
	  sscanf (optp->answers[count], "%f", &prob[count]);
	  count ++;
	}
      if (count != no_transforms)
	G_fatal_error ("Wrong number of probabilities ");
    }

  /* ...else calculate them */

  else
    {
      sum_determinants = 0;
      for (t = 0; t < no_transforms; t++)
	{
	  determinant [t] = fabs((fn[a][t] * fn[d][t]) -
				 (fn[b][t] * fn [c][t]));
	  if (determinant [t] < kSMALL_PROB)
	    determinant [t] = kSMALL_PROB;
	  sum_determinants += determinant[t];
	}
      
      for (t = 0; t < no_transforms; t++)
	{
	  /* Calculate probabilities */
	 prob [t] = (determinant [t] / sum_determinants);
	}
    }


  /* Print parameters to stdout */

  if (!silent->answer)
    {
      if (output->answer != NULL)
	printf ("\n\nParameters for '%s' are:", output->answer);
      else
	printf ("\n\nParameters are:");
      printf ("\n t        a        b        c        d        e        f        p");
      for (t = 0; t < no_transforms; t++)
	{
	  printf ("\n%2d %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f",
		  t+1, fn[a][t], fn[b][t], fn[c][t], fn[d][t],
		  fn[e][t], fn[f][t],
		  prob [t]);
	  fflush (stdout);
	}
  printf ("\n");
    }


  /***********************************************************************/
  /* Initialise variables used in ifs                                    */
  /***********************************************************************/

  /* Initialize region-specific constants */

  if (width->answer == NULL)
    {
      left = 30;
      if (ncols <= nrows)
	w = (float) ncols;
      else
	w = (float) nrows;
    }
  else 
    w = (float) wdth;

  /* Scale translations e and f according to region */
  
  for (t = 0; t < no_transforms; t++)
    {
     fn[e][t] *= w;
     fn[f][t] *= w;
    }


  /* Set first game point */
  x = fn[e][0];
  y = 0;


  /* Calculate cumulative probabilities */
  
  t= 0;
  cum_prob [t] = prob [t];
  if (no_transforms > 1)
    {
      for (t = 1; t < no_transforms; t++)
	{
	  cum_prob [t] = cum_prob [t - 1] + prob [t];
	}
    }
  t = no_transforms - 1;
  if (((cum_prob [t]) > 1.01) ||
      ((cum_prob [t]) < 0.99))
    G_fatal_error ("Transformation probabilities do not sum to 1 ");

  
  /***********************************************************************/
  /* Generate fractal                                                    */
  /***********************************************************************/

  srand (rand_seed);
  if (!terse->answer)
    fprintf (stderr, "\nIteration =         ");
  val = 1;
  for (i = 1; i <= its; i++)
    {
      r = ((float) rand ()) / RAND_MAX;
      for (t = 0; t < no_transforms; t++)
	{
	  if (r <= cum_prob[t])
	    {
	      xn = fn[a][t] * x + fn[b][t] * y + fn[e][t];
	      yn = fn[c][t] * x + fn[d][t] * y + fn[f][t];
	      break;
	    }
	}

      /* Adjust y because GRASS arrays have origin at top */
  
      ploty = array_south - (int) y;


      /* Check whether point falls in current region */

      if ((ploty >= 0 && ploty <= nrows) && (x <= ncols))
	{

	  /* If output map required */
      
	  if (output->answer != NULL)
	    segment_put (&outSEG, (char*) &val, ploty, (int) x);
	  
	  /* If display on monitor required */
	  
	  if (display->answer)
	    {
	      screenx [0] = (int) D_a_to_d_col ((double) x);
	      screeny [0] = (int) D_a_to_d_row ((double) ploty);
	      R_polydots_abs (screenx, screeny, 1);
	    }
	}
      if (!terse->answer)
	fprintf (stderr, "\b\b\b\b\b\b\b\b%8ld", i);
      
      x = xn;
      y = yn;
    }
  

  /***********************************************************************/
  /* Tidy up                                                             */
  /***********************************************************************/

  /* If necessary write output map*/
  
  if (output->answer != NULL)
    {
      if (!terse->answer)
	fprintf (stderr, "\nSaving results in raster map '%s' ", 
		 output->answer);
      segment_flush (&outSEG);
      for (row = 0; row < nrows; row ++)
	{
	  segment_get_row (&outSEG, (char*) cell, row);
	  if (G_put_map_row (out_map_fd, cell) < 0)
	    G_fatal_error ("Failed to convert output submatrices ");
	}
      segment_release (&outSEG); close (out_seg_fd);
      unlink (out_seg_name); G_close_cell (out_map_fd);
    }


  /* If display on monitor clean up graphics */
  
  if (display->answer)
    {
      R_stabilize ();
      R_close_driver();
    }

  /* Create history support file for output map */


  if (!output->answer == NULL)
    {
      G_short_history (output->answer, "raster", &history);
      history.edlinecnt = no_transforms + 2;
      sprintf (history.edhist [0], "Parameters are:");
      sprintf (history.edhist [1], "t        a        b        c        d        e        f        p");
      for (t = 2; t < (no_transforms + 2); t++)
	{
	  sprintf (history.edhist [t], 
		   "\n%2d %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f %8.4f",
		   t+1, fn[a][t-2], fn[b][t-2], fn[c][t-2], fn[d][t-2],
		   fn[e][t-2], fn[f][t-2],
		   prob [t-2]);
	}
      if (G_write_history (output->answer, &history) == -1)
	G_warning ("Failed to write history file ");
    }

  if (!terse->answer)
    fprintf (stderr, "\n\nJob finished\n");
}
